/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 * 
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Microsoft Exchange Calendar Provider code.
 *
 * The Initial Developer of the Original Code is
 *   Andrea Bittau <a.bittau@cs.ucl.ac.uk>, University College London
 * Portions created by the Initial Developer are Copyright (C) 2009
 * the Initial Developer. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *      
 * ***** END LICENSE BLOCK ***** */

// URI format for setting up calendar:
// http://myemail@address.com@auto for auto discovery
//
// https://myemail@address.com@myexchangeserver.com/EWS/Exchange.asmx for
// hardcoded URL.
//
// myemail@address.com is the username to log in the Exchange server.

const _mst = "http://schemas.microsoft.com/exchange/services/2006/types";
const _msm = "http://schemas.microsoft.com/exchange/services/2006/messages";

function convDate(aDate)
{
	var d = aDate.clone();

	d.isDate = false;
	return _f["toRFC3339"](d);
}

function calExchange()
{
	this.initProviderBase();
	this.mReqQueue = new Array();
	this.mRunning = false;
}

calExchange.prototype = {
	__proto__: _proto,

	QueryInterface: function(aIID)
	{
		return doQueryInterface(this, calExchange.prototype, aIID,
					[Components.interfaces.calICalendarProvider,
					 Components.interfaces.calISchedulingSupport,
					 Components.interfaces.calIFreeBusyProvider]);
	},

	get type()
	{
		return "exchange";
	},

	set uri(aUri)
	{
		this.mUri = aUri;

		this.mUser = decodeURIComponent(aUri.username);

		this.sdbg("User:" + this.mUser + "\n");

		if (aUri.host != "auto") {
			this.mServer = decodeURIComponent(aUri.spec);
			this.sdbg("Server: " + this.mServer + "\n");
		}

		return aUri;
	},

	get uri()
	{
		return this.mUri;
	},

	getProperty: function(aName)
	{
		switch (aName) {
		case "capabilities.tasks.supported":
			return false;
		}

		return this.__proto__.__proto__.getProperty.apply(this,
								  arguments);
	},

	get canRefresh()
	{
		return true;
	},

	canNotify: function(aMethod, aItem)
	{
		return true;
	},

	init: function()
	{
		if (this.mInit != 0)
			return;

		this.mInit = 1;
		this.sdbg("Init\n");

		getFreeBusyService().addProvider(this);

		var tmp = this;

		let timerCallback = {
			notify: function setTimeout_notify() {
				tmp.doInit();
			}
		};

		let timer = Components.classes["@mozilla.org/timer;1"]
				.createInstance(Components.interfaces.nsITimer);

		timer.initWithCallback(timerCallback, 0, timer.TYPE_ONE_SHOT);
	},

	doInit: function()
	{
		this.sdbg("do init\n");
		this.getAccount();
		this.newXmlReq();

		if (this.mServer == null)
			this.autodiscover();

		this.mInit = 2;
		this.kick();
		this.sdbg("do init done\n");
	},

	getItem: function (aId, aListener)
	{
		var data = makeFindItemUID(aId);

		var req = new calExchangeRequest(this, data,
						 "findItemReply");

		req.e4x = true;
		req.mListener = aListener;
		req.mOp = Components.interfaces.calIOperationListener.GET;

		this.queueRequest(req);
	},

	getItems: function(aItemFilter, aCount, aRangeStart, aRangeEnd,
			   aListener)
	{
		this.init();

		var wantInvitations = 
			((aItemFilter & Components.interfaces.calICalendar
		  		.ITEM_FILTER_REQUEST_NEEDS_ACTION) != 0);

		if (!(aItemFilter & Components.interfaces
				.calICalendar.ITEM_FILTER_TYPE_EVENT)
                    && !wantInvitations) {
			this.completeGetRequest(null, aListener);
			return;
		}

		var data = makeFindItem(aItemFilter, aCount, aRangeStart, aRangeEnd);

		var req = new calExchangeRequest(this, data,
						 "findItemReply");

		req.e4x = true;
		req.mListener = aListener;
		req.mOp = Components.interfaces.calIOperationListener.GET;
		req.mInvite = wantInvitations;
		req.mItemFilter = aItemFilter;
		req.mRangeStart = aRangeStart;
		req.mRangeEnd = aRangeEnd;

		this.queueRequest(req);
	},

	completeGetRequest: function(aID, aListener)
	{
		if (!aID) {
			aID = null;
		}

		if (aListener == null) {
			aListener = this.mReq.mListener;
		}

		this.notifyOperationComplete(aListener,
		      Components.results.NS_OK,
		      Components.interfaces.calIOperationListener.GET,
		      aID,
		      null);
	},

	queueRequest: function(aReq, aFront)
	{
		this.sdbg("doRequest\n");
		if (aFront) {
			this.mReqQueue.unshift(aReq);
		} else {
			this.mReqQueue.push(aReq);
		}
		this.kick();
	},

	kick: function()
	{
		if (this.mInit != 2)
			return;

		if (this.mRunning
		    || this.mReqQueue.length == 0)
			return;

		this.mRunning = true;

		var req = this.mReqQueue.shift();

		this.doRequest(req);
	},

	autodiscover: function()
	{
		// XXX should do:
		// 1) domain/autodiscover/autodiscover.xml
		// 2) autodiscover.domain/autodiscover/autodiscover.xml

		var email;
		var identity = this.getProperty('imip.identity');
		if (identity) {
			email = identity.QueryInterface(Components.interfaces.nsIMsgIdentity).email;
		}
		email = email || decodeURIComponent(mUri.username);

		this.sdbg("autodiscover\n");

		var parts = email.split("@");
		var url = "https://autodiscover."
			  + parts[1] 
			  + "/autodiscover/autodiscover.xml";

		this.setupReq(url);

		var data = makeAutodiscover(email);
		var req  = new calExchangeRequest(this, data,
						  "autodiscoverReply");

		this.mReqQueue.unshift(req);
		this.kick();
	},

	doRequest: function(aReq)
	{
		this.mReq = aReq;

		if (this.mServer != null)
			this.setupReq(this.mServer);

		this.sdbg("Sending: " + aReq.mData + "\n");
		this.mXmlReq.send(aReq.mData);
	},

	reply: function()
	{
		var req = this.mXmlReq;

		if (req.readyState != 4)
			return;

		if (req.status == 302) {
			var loc = this.mHttp.getResponseHeader("Location");

			this.sdbg("Redirect: " + loc + "\n");

			// XXX pheer loops.
			this.mXmlReq.abort();
			this.newXmlReq();
			this.setupReq(loc);
			this.doRequest(this.mReq);

			return;
		}

		this.sdbg("reply\n");

		var xml = req.responseText; // bug 270553
		xml = xml.replace(/^<\?xml\s+version\s*=\s*(["'])[^\1]+\1[^?]*\?>/, ""); // bug 336551
		xml = new XML(xml);

		this.sdbg(String(xml) + "\n");

		this.mAuthFail = 0;
		this.mRunning  = false;

		var resp;
		if (this.mReq.e4x) {
			resp = xml;
		} else {
			resp = req.responseXML;
		}

		if (!this.isError(req))	
			this[this.mReq.mCb](resp);

		this.kick();
	},

	isError: function(aReq)
	{
		if (aReq.status != 200) {
			if (aReq.status == 401) {
				this.sdbg(aReq.responseText + "\n");
				this.discardAccount();
			}
			// XXX parse it
			this.fail(aReq.responseText);
			return true;
		}

		var aReply = aReq.responseXML;

		var err = getXml(aReply, "ResponseCode", _msm);

		if (err == "" || err == "NoError")
			return false;

		var msg = getXml(aReply, "MessageText", _msm); 

		this.sdbg("Error: " + err + " msg: " + msg + "\n");

		this.fail(msg);
		return true;
	},

	newXmlReq: function()
	{
		this.mXmlReq = 
		  Components.classes ["@mozilla.org/xmlextras/xmlhttprequest;1"]
		   .createInstance(Components.interfaces.nsIJSXMLHttpRequest);
	},

	fail: function(aMsg)
	{
		this.sdbg("fail status: " + this.mXmlReq.status + " " + aMsg + "\n");

		var req  = this.mReq;

		if (req.mListener) {
			// XXX force prompt
			const calIErrors = Components.interfaces.calIErrors;
			var code = calIErrors.MODIFICATION_FAILED;

			if (req.mAvail != null) {
				req.mListener.onResult(null, null);

				var xml = this.mXmlReq.responseXML;
				var err = getXml(xml, "ResponseCode", _msm);

				// XXX should avoid the second one
				if (err == "ErrorMailRecipientNotFound"
				    || err == "ErrorInvalidSmtpAddress")
					return;
			} else {
				req.mListener.onOperationComplete(this, code, req.mOp, null, aMsg);
			}
		}

		this.mObservers.notify("onError", [this, code, aMsg]);

		return;
	},

	setupReq: function(aUrl)
	{
		this.sdbg("setup req to " + aUrl + " user " + this.mUser + "\n");
		this.mXmlReq.open("POST", aUrl, true, this.mUser, this.mPass);

		var tmp = this;

		this.mXmlReq.onreadystatechange = function() {
			tmp.reply();
		}

		this.mXmlReq.overrideMimeType('text/xml');
		this.mXmlReq.setRequestHeader("Content-Type", "text/xml");

		this.mXmlReq.channel.notificationCallbacks = this;

		var httpChannel = this.mXmlReq.channel.QueryInterface(
			Components.interfaces.nsIHttpChannel);

		// XXX we want to preserve POST across 302 redirects
		httpChannel.redirectionLimit = 0;

		this.mHttp = httpChannel;
	},

	autodiscoverReply: function(aReply)
	{
		this.sdbg("autodiscoverreply");
		this.mServer = getXml(aReply, "ASUrl");

		ASSERT(this.mServer != "");

		this.sdbg("Autodiscover server: " + this.mServer + "\n");
	},

	findItemReply: function(aReply)
	{
		// FindItem does not return body.  Calendar's GetItem never
		// seems to be called, so we gotta be proactive and call
		// Exchange's getItem to retrieve body.

		var ids = readFindItemReply(aReply);

		this.sdbg("findItemReply: " + ids.length + "\n");

		if (ids.length == 0) {
			this.completeGetRequest();
			return;
		}

		var data = makeGetItem(ids, this.mReq.mItemFilter);
		var req  = new calExchangeRequest(this, data, "getItemReply");
		req.mListener = this.mReq.mListener;
		req.mOp = this.mReq.mOp;
		req.mInvite = this.mReq.mInvite;
		req.mItemFilter = this.mReq.mItemFilter;
		req.mRangeStart = this.mReq.mRangeStart;
		req.mRangeEnd = this.mReq.mRangeEnd;
		req.e4x = true;

		this.queueRequest(req, true);
	},

	getItemReply: function(aReply)
	{
		this.sdbg("getItemReply\n");

		var el = this.mReq.mInvite ? "MeetingRequest" : "CalendarItem";

		var items = readGetItemReply(this, aReply, this.mReq.mItemFilter, this.mReq.mRangeStart, this.mReq.mRangeEnd);
		for each (var it in items) {
			it.setProperty("X-SOR-INVITE", this.mReq.mInvite);
		}

		var listener = this.mReq.mListener;

		listener.onGetResult(this,
				     Components.results.NS_OK,
				     Components.interfaces.calIEvent,
				     null,
				     items.length,
				     items);

		this.completeGetRequest();
	},

	modifyItem: function(aNewItem, aOldItem, aListener)
	{
		this.sdbg("modifyItem\n");

		/* just in case we need to modify it */
		aNewItem = aNewItem.clone();

		var data;
		var del = false;
		var flags = {};

		var invite = aNewItem.getProperty("X-SOR-INVITE") || this.isInvitation(aNewItem);

		if (invite === true) {
			/* XXX only come here if the participation status changed */
			var me = this.getInvitedAttendee(aNewItem);

			if (me.participationStatus == "DECLINED")
				del = true;

			this.sdbg("PS: " + me.participationStatus + "\n");

			data = makeMeetingResponse(aNewItem, me.participationStatus);

			// XXX
			var a = this.getInvitedAttendee(aOldItem)
					.participationStatus;
			var b = me.participationStatus;
			if (((a == b) 
				&& (a == "ACCEPTED" || a == "NEEDS-ACTION"))
			    || ((a == "TENTATIVE") && (b == "NEEDS-ACTION")))
				data = "";
		} else
			data = makeUpdateItem(aNewItem, aOldItem, flags);

		if (data == "") {
			this.completeUpdateRequest(aListener, aNewItem,
						   aOldItem);

			return;
		}

		var req = new calExchangeRequest(this, data,
						 flags.createExceptions
						     ? "updateItem_updateExceptions"
						     : "updateItemReply");

		req.mListener = aListener;
		req.mNewItem  = aNewItem;
		req.mOldItem  = aOldItem;
		req.mOp = Components.interfaces.calIOperationListener.MODIFY;
		req.e4x = true;
		req.mDel = del;
		req.mFlags = flags;

		this.queueRequest(req);
	},

	/*
	 * When changing the recurrence of a master, we will lose all
	 * exception / deleted information.  This function sets out
	 * to recreate those.
	 *
	 * We will operate in two steps, each as separate request:
	 * - first we delete occurrences if necessary
	 * - then we modify all exceptions
	 */
	updateItem_updateExceptions: function(aReply)
	{
		var oreq = this.mReq;

		/* build a GetItem requests to sync what Exchange did with our update */

		var newitem = oreq.mNewItem.clone();

		var newid = aReply..t::ItemId[0];
		if (newid) {
			newitem.setProperty("SEQUENCE", newid.@ChangeKey.toString());
		} else {
			/* Bad luck.  Better ask for the latest Id */
			var nreq = oreq;
			nreq.mData = makeGetItemId(newitem);
			this.queueRequest(nreq, true);
			return;
		}

		let nolditem = newitem.clone();
		//nolditem.recurrenceInfo = nolditem.recurrenceInfo.clone();

		let havedel = false;
		let havepos = false;
		let ritems = [];
		for each (let ritem in newitem.recurrenceInfo.getRecurrenceItems({})) {
			if (!cal.calInstanceOf(ritem, Components.interfaces.calIRecurrenceDate)) {
				ritems.push(ritem);
				continue;
			}

			if (ritem.isNegative) {
				havedel = true;
			}
		}

		if (newitem.recurrenceInfo.getExceptionIds({}).length > 0) {
			havepos = true;
		}

		nolditem.recurrenceInfo.setRecurrenceItems(ritems.length, ritems);

		var xreq = null;
		var flags = oreq.mFlags;

		if (!oreq.mFlags.didDelete && havedel) {
			xreq = checkDeletedOccurrences(newitem, nolditem);
			flags.didDelete = true;
			flags.createExceptions = havepos;
		} else {
			let upds = new XMLList();
			for each (let exid in newitem.recurrenceInfo.getExceptionIds({})) {
				nolditem.recurrenceInfo.removeExceptionFor(exid);
				let upoitem = nolditem.recurrenceInfo.getOccurrenceFor(exid);
				let upnitem = newitem.recurrenceInfo.getOccurrenceFor(exid);
				let upd = makeUpdateOneItem(upnitem, upoitem);
				upds += upd;
			}

			xreq = makeUpdateItem2(newitem, upds);
			flags.createExceptions = false;
		}

		var data = makeSoapMessage(xreq);

		var req = new calExchangeRequest(this, data,
						 flags.createExceptions
						     ? "updateItem_updateExceptions"
						     : "updateItemReply");

		req.mListener = oreq.mListener;
		req.mNewItem  = oreq.mNewItem;
		req.mOldItem  = oreq.mOldItem;
		req.mOp = Components.interfaces.calIOperationListener.MODIFY;
		req.e4x = true;
		req.mDel = oreq.mDel;
		req.mFlags = flags;

		this.queueRequest(req, true);
	},

	updateItemReply: function(aReply)
	{
		this.sdbg("updateItemReply\n");

		var oreq = this.mReq;

		/* build a GetItem requests to sync what Exchange did with our update */

		var newitem = oreq.mNewItem.clone();

		var newid = aReply..t::ItemId[0];
		if (newid) {
			/*
			 * We have to be careful.  If we (re-)set the id of the
			 * recurring master, the calendar backend will automatically
			 * overwrite all children's ids.
			 */
			newitem.setProperty("X-ITEMID", newid.@Id.toString());
			newitem.setProperty("SEQUENCE", newid.@ChangeKey.toString());
		}

		var ids = [];
		if (!newitem.parentItem.recurrenceInfo) {
			ids.push({Id: newitem.getProperty("X-ITEMID")});
		} else {
			/*
			 * We have to list all the exceptions of our master, or
			 * otherwise the GetItem will reconstruct the master
			 * incompletely.
			 *
			 * If we modify an exception, make sure that we
			 * ask for it as first item, so we can notify the
			 * listener with the right item lateron.
			 */
			if (newitem.parentItem != newitem) {
				ids.push({Id: newitem.getProperty("X-ITEMID"), type: "Exception"});
			}

			var recurrenceInfo = newitem.parentItem.recurrenceInfo;
			for each (let id in recurrenceInfo.getExceptionIds({})) {
				if (newitem.recurrenceId && id.compare(newitem.recurrenceId) == 0) {
					/* we added this one already */
					continue;
				}
				let exitem = recurrenceInfo.getExceptionFor(id);
				ids.push({Id: exitem.getProperty("X-ITEMID"), type: "Exception"});
			}

			/*
			 * Seems this was an update deleting an occurrence, and there
			 * is no other exception.  We'll have to ask for the master
			 * then.
			 */
			if (ids.length == 0) {
				ids.push({Id: newitem.parentItem.getProperty("X-ITEMID")});
			}
		}
		var data = makeGetItem(ids);

		var req = new calExchangeRequest(this, data, "resyncUpdateItemReply");

		req.mListener = oreq.mListener;
		req.mNewItem  = oreq.mNewItem;
		req.mOldItem  = oreq.mOldItem;
		req.mOp = Components.interfaces.calIOperationListener.MODIFY;
		req.e4x = true;
		req.mDel = oreq.mDel;

		this.queueRequest(req, true);
	},

	resyncUpdateItemReply: function(aReply)
	{
		this.sdbg("resyncUpdateItemReply\n");
		var items = readGetItemReply(this, aReply);

		var listener = this.mReq.mListener;
		var item = items[0];

		this.completeUpdateRequest(listener, item, this.mReq.mOldItem);

		if (this.mReq.mDel)
			this.mObservers.notify("onDeleteItem", [item]);
	},

	completeUpdateRequest: function(aListener, aNewItem, aOldItem)
	{
		this.notifyOperationComplete(aListener,
					     Components.results.NS_OK,
					     Components.interfaces.calIOperationListener.MODIFY,
					     aNewItem.id,
					     aNewItem);

		if (aNewItem.parentItem != aNewItem) {
			var master = aNewItem.parentItem.clone();

			master.recurrenceInfo.modifyException(aNewItem, false);
			aNewItem = master;
			aOldItem = aOldItem.parentItem;
		}

		this.mObservers.notify("onModifyItem",
				       [aNewItem, aOldItem]);
	},

	deleteItem: function(aItem, aListener)
	{
		this.sdbg("deleting item\n");

		var data = makeDeleteItem(aItem);

		var req = new calExchangeRequest(this, data,
						 "deleteItemReply");

		req.mListener = aListener;
		req.mItem     = aItem;
		req.mOp = Components.interfaces.calIOperationListener.DELETE;

		this.queueRequest(req);
	},

	deleteItemReply: function(aReply)
	{
		this.sdbg("deleteItemReply\n");

		var item     = this.mReq.mItem;
		var listener = this.mReq.mListener;

		listener.onOperationComplete(this,
                      Components.results.NS_OK,
                      Components.interfaces.calIOperationListener.DELETE,
                      item.id,
                      item);

		this.mObservers.notify("onDeleteItem", [item]);
	},

	addItem: function(aItem, aListener)
	{
		return this.adoptItem(aItem.clone(), aListener);
	},

	adoptItem: function(aItem, aListener)
	{
		this.sdbg("adding item\n");

		if (!aItem.id) {
			aItem.id = cal.getUUID();
		}

		var data = makeCreateItem(aItem);

		var req = new calExchangeRequest(this, data,
						 "createItemReply");

		req.e4x = true;
		req.mListener = aListener;
		req.mItem     = aItem;
		req.mOp = Components.interfaces.calIOperationListener.ADD;

		this.queueRequest(req);
	},

	createItemReply: function(aReply)
	{
		this.sdbg("createItemReply\n");

		var item     = this.mReq.mItem;
		var listener = this.mReq.mListener;

		var id = aReply..t::ItemId[0];
		item.setProperty("X-ITEMID", id.@Id.toString());
		item.setProperty("SEQUENCE", id.@ChangeKey.toString());

		listener.onOperationComplete(this,
                      Components.results.NS_OK,
                      Components.interfaces.calIOperationListener.ADD,
                      item.id,
                      item);

		this.mObservers.notify("onAddItem", [item]);
	},

	getFreeBusyIntervals: function(aCalId, aRangeStart, aRangeEnd,
				       aBusyTypes, aListener)
	{
		this.sdbg("getFreeBusyIntervals: " + aCalId + "\n");

		if (aCalId.indexOf("@") < 0 || aCalId.indexOf(".") < 0) {
            		// No valid email, screw it
			aListener.onResult(null, null);
			return;
		}

		var data = makeGetUserAvailability(
					aCalId.replace(/^MAILTO:/, ""),
					aRangeStart, aRangeEnd);

		var req = new calExchangeRequest(this, data,
						 "UserAvailabilityReply");

		req.mListener = aListener;
		req.mAvail    = true;
		req.mCalId    = aCalId;

		this.queueRequest(req);
	},

	UserAvailabilityReply: function(aReply)
	{
		this.sdbg("UserAvailabilityReply\n");

		var items = new Array();

		var ci = aReply.getElementsByTagNameNS(_mst, "CalendarEvent");
		for (i = 0; i < ci.length; i++) {
			var item = this.doAvailability(ci[i]);
			items.push(item);
		}

		this.mReq.mListener.onResult(null, items);
	},

	doAvailability: function(aCi)
	{
		const x = Components.interfaces.calIFreeBusyInterval;

		const types = {
			"Free" 		: x.FREE,
			"Tentative"	: x.BUSY_TENTATIVE,
			"Busy"		: x.BUSY,
			"OOF"		: x.BUSY_UNAVAILABLE,
			"NoData"	: x.UNKNOWN
		};

		/* We ask Exchange to report the time in UTC */

		var start = _f["fromRFC3339"](getXml(aCi, "StartTime", _mst), cal.UTC());
		var end   = _f["fromRFC3339"](getXml(aCi, "EndTime", _mst), cal.UTC());
		var type  = types[getXml(aCi, "BusyType", _mst)];

		return new _f["FreeBusyInterval"](this.mReq.mCalId, type,
					       	  start, end);
	},

	// XXX seem to get auth prompts for NTLM.  Suppress them here.
	getInterface: function(iid)
	{
		this.sdbg("getInterface\n");

		if (iid.equals(Components.interfaces.nsIAuthPrompt)) {
			if (this.mAuthFail++ > 2) {
				this.mAuthFail = 0;
				if (!this.discardAccount()) {
					return fakePrompt(this, false);
				}
			}

			return fakePrompt(this);
		} 

		this.sdbg("IID: " + iid + "\n");

//		throw Components.results.NS_ERROR_NO_INTERFACE;
		return null;
	},

	getAccount: function()
	{
		var save = true;

		this.sdbg("getAccount\n");

                var username = { value: this.mUser };
                var password = { value: this.mPass };
                var persist  = { value: false };
		var title = "Microsoft Exchange";
		var realm = this.mUri.prePath;

		var got = _f["passwordManagerGet"]
				(this.mUser, password, realm, realm);

		if (got) {
			this.sdbg("USER: " + this.mUser + "\n");
			this.mPass = password.value;
			return true;
		}

		var ok = _f["getCredentials"](title, realm, username, password, persist);

		if (!ok) {
			this.mUser = null;
			this.mPass = null;
			return false;
		}

		this.mUser = username.value;
		this.mPass = password.value;

		this.sdbg("USER: " + this.mUser + "\n");

		//this.sdbg("PASS: " + this.mPass + "\n");

		if (persist.value) {
			_f["passwordManagerSave"](this.mUser, this.mPass,
						  realm, realm);
		}

		return true;
	},

	discardAccount: function()
	{
		if (!this.mUser || !this.mPass) {
			return false;
		}

		var realm = this.mUri.prePath;

		this.sdbg("discarding password info on " + realm + " / " + this.mUser + "\n");

		_f["passwordManagerRemove"](this.mUser, realm, realm);
		return this.getAccount();
	},

	sdbg: function(aText)
	{
		var id = this.id;

		if (id === null || id === undefined) {
			// we will confuse the settings if we query the name too early
			sdbg("unknown " + this + ": " + aText);
			return;
		}

		sdbg(this.id + "/" + this.name + ": " + aText);
	},

	mReqQueue: null,
	mServer: null,
	mUri: null,
	mRunning: null,
	mUser: null,
	mPass: null,
	mReq: null,
	mXmlReq: null,
	mInit: 0,
	mAuthFail: 0
};

function calExchangeRequest(aCal, aData, aCb)
{
	this.wrappedJSObject = this;
	this.mCal  = aCal;
	this.mData = aData;
	this.mCb   = aCb;
}

calExchangeRequest.prototype = {
	QueryInterface: function(aIID) {
		return doQueryInterface(this,
					calExchangeRequest.prototype,
					aIID,
					null,
					g_classInfo["calExchangeRequest"]);
	},

	mCal: null,
	mData: null,
	mCb: null,
	mListener: null
};

function sendMail(aItem)
{
	var inv = aItem.getProperty("X-MOZ-SEND-INVITATIONS");

	if (inv == "TRUE" || true)
		return "SendToAllAndSaveCopy";

	return "SendToNone";
}

function getXml(aXml, aName, aNs)
{
	var items;

	if (aNs == null)
		items = aXml.getElementsByTagName(aName);
	else
		items = aXml.getElementsByTagNameNS(aNs, aName);
	
	if (items.length == 0)
		return "";

	return items[0].textContent;
}

function doXPath(aXml, aXPath)
{
	var doc = aXml.ownerDocument;

	var res = function(aPref) {
		var ns = {
			't' : _mst
		};

		return ns[aPref] || null;
	}

	return doc.evaluate(aXPath, aXml, res,
		     Components.interfaces.nsIDOMXPathResult.ANY_TYPE, null);
}

function fakePrompt(aCal, aSuccess)
{
	if (aSuccess === undefined) {
		aSuccess = true;
	}

	sdbg("making prompt: " + aSuccess + "\n");

	var p = {
		promptUsernameAndPassword: function(dialogTitle,
						    text,
						    passwordRealm,
						    savePassword,
						    user,
						    pwd)
		{
			user.value = aCal.mUser;
			pwd.value  = aCal.mPass;

			return aSuccess;
		}
	}

	return p;
}

function getDate(aRange)
{
	// Requesting only a DATE returns items based on UTC. Therefore,
	// we make sure both start and end dates include a time and
	// timezone. This may not quite be what was requested, but I'd
	// say its a shortcoming of rfc3339.  
	//
	// [ripped comment from gdata - like many other things ;D]
	if (aRange) {
		aRange = aRange.clone();
		aRange.isDate = false;
	}

	return _f["toRFC3339"](aRange);
}

function sdbg(msg)
{
	//return;
	cal.LOG(msg);
}
